import argparse
import serial

from mavsniff.utils.log import logger
from mavsniff.replay import Replay

class ReplayCommand:

    def run(self, args: argparse.Namespace):
        logger.debug(f"opening {args.file} for reading")

        file = None
        try:
            file = open(args.file, 'rb')
        except IOError:
            logger.error(f"failed to open {args.file} for reading")
            raise

        device = None
        try:
            device = serial.serial_for_url(args.device)
        except serial.SerialException:
            logger.error(f"failed to open {args.device} for writing")
            raise

        try:
            replayed = Replay(file, device).run(limit=args.limit)
            logger.info(f"replayed {replayed} valid MAVLink packets")
        finally:
            file.close()
            device.close()



import sys
import argparse
import logging

from .commands import CaptureCommand, ReplayCommand
from .utils.log import logger

commands = {
    "capture": CaptureCommand(),
    "replay": ReplayCommand(),
}

parser = argparse.ArgumentParser(
            prog='mavsniff',
            description='Capture and replay mavlink packets',
            epilog="You can see available serial devices with 'python -m serial.tools.list_ports -v'",
            )
parser.add_argument('command', choices=commands.keys())
parser.add_argument("--file", "-f", required=True, dest="file", help="pcap file to read from or write to")
parser.add_argument("--device", "-d", required=True, dest="device", help="device (/dev/ttyUSB0 or /dev/ttyS0 on linux, COM1 on windows or simply loop:// for testing)")
parser.add_argument("--limit", "-l", dest="limit", type=int, default=-1, help="limit the number of read/written packets (default -1 unlimited)")
parser.add_argument("--verbose", "-v", dest="debug", action='store_true', help="enable debug logging")

if len(sys.argv) == 1:
    parser.print_help()
    sys.exit(1)

args = parser.parse_args(sys.argv[1:])

logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

try:
    cmd = commands.get(args.command)
    cmd.run(args)
except Exception as e:
    logger.error(e)
    sys.exit(1)


import argparse
import serial

from mavsniff.utils.log import logger
from mavsniff.capture import Capture

class CaptureCommand:

    def run(self, args: argparse.Namespace):
        logger.debug(f"opening {args.file} for writing")

        file = None
        try:
            file = open(args.file, 'wb')
        except IOError:
            logger.error(f"failed to open {args.file} for writing")
            raise

        device = None
        try:
            device = serial.serial_for_url(args.device)
        except serial.SerialException:
            logger.error(f"failed to open {args.device} for reading")
            raise

        try:
            captured = Capture(device, file).run(limit=args.limit)
            logger.info(f"captured {captured} valid MAVLink packets")
            raise
        finally:
            device.close()
            file.close()




import serial
import pcapng
import time
import io

from mavsniff.utils.mav import MavSerial
from mavsniff.utils.log import logger


INTERFACE_MAGIC = 0x00000001
PACKET_MAGIC = 0x00000006
SECTION_MAGIC = 0x0A0D0D0A


class Replay:
    def __init__(self, file: io.BytesIO, device: serial.Serial, mavlink_version=2):
        self.scanner = pcapng.FileScanner(file)
        self.device = MavSerial(device, mavlink_version=mavlink_version)

    def run(self, limit=-1) -> int:
        """Replay a PCAPNG file to a device"""
        reader = iter(self.scanner)
        section_header = next(reader)
        if section_header.magic_number != SECTION_MAGIC:
            raise ValueError("invalid PCAPNG file - does not start with section header")

        interface_description = next(reader)
        if interface_description.magic_number != INTERFACE_MAGIC:
            raise ValueError("invalid PCAPNG file - does not have interface header")

        # Resolution is handled in the mavlink library - timestamp is in seconds
        # resolution seems to be constant for all packets in a file
        # self.resolution_ts = interface_description.timestamp_resolution
        self.last_packet_ts = time.time()
        self.last_sent_ts = 0.0
        packets_written = 0

        while True:
            try:
                packet = next(reader)
                if packet is None:
                    logger.debug("no more packets to read")
                    break
                if packet.magic_number != PACKET_MAGIC:
                    logger.debug(f"discarding non-data packet {packet}")
                    continue
                self._send_in_timely_manner(packet); packets_written += 1
                if limit > 0 and packets_written >= limit:
                    logger.debug(f"reached packet limit of {limit}")
                    raise StopIteration()
            except StopIteration:
                break

        return packets_written


    def _send_in_timely_manner(self, packet):
        """Replay a packet to the device"""
        packet_ts_delta = packet.timestamp - self.last_packet_ts
        since_last_sent = time.time() - self.last_sent_ts
        sleep_time = (packet_ts_delta - since_last_sent)
        if sleep_time > 0.000001:
            logger.debug(f"sleeping for {sleep_time} seconds")
            time.sleep(sleep_time)
        self.device.write(packet.packet_data)
        self.last_sent_ts = time.time()
        self.last_packet_ts = packet.timestamp



import pcapng
import serial
import io
import time
import struct

import signal
import time

from mavsniff.utils.mav import MavSerial
from mavsniff.utils.log import logger


class Capture:
    """Capture reads Mavlink messages from a device and store them into a PCAPNG file"""
    def __init__(self, device: serial.Serial, file: io.BytesIO, mavlink_version=2):
        self.file = file
        self.done = False
        self.interface_id=0x00
        self.device = MavSerial(device, mavlink_version=mavlink_version)
        self.sbh = pcapng.blocks.SectionHeader(msgid=0, endianness="<", options={
            'shb_userappl': 'mavsniff',
        })
        self.sbh.register_interface(pcapng.blocks.InterfaceDescription(msdgid=0x01, endianness="<", interface_id=self.interface_id, section=self.sbh, options={
            'if_name': self.device.name,
            'if_txspeed': self.device.baudrate,
            'if_rxspeed': self.device.baudrate,
            'if_tsresol': struct.pack('<B', 6), # negative power of 10
            # should we deal with timestamp resolution?
        }))
        signal.signal(signal.SIGINT, self.stop)

    def run(self, limit=-1, limit_invalid_packets=-1) -> int:
        """Store Mavlink messages into a PCAPNG file"""
        self.writer = pcapng.FileWriter(self.file, self.sbh)
        self.done = False
        received = 0
        parse_errors = 0
        proceed = lambda: not self.done and (limit < 0 or received < limit)
        while proceed():
            try:
                msg = self.device.recv_msg()
                parse_errors = 0
            except self.device.ParseError:
                parse_errors += 1
                if limit_invalid_packets > 0 and parse_errors > limit_invalid_packets:
                    raise RuntimeError("Too many invalid packets in a row")
                continue
            except serial.SerialException:
                logger.info("serial line closed")
                break
            if msg is None:
                logger.debug("dropping empty message")
                continue
            if msg.get_type() == 'BAD_DATA':
                logger.debug(f"dropping invalid message {msg}")
                continue
            self._write_packet(msg.pack(self.device.mav))
            received += 1

        return received

    def _write_packet(self, packet_bytes):
        """Write packet to the device"""
        now_us = time.time_ns() // 1000
        self.writer.write_block(pcapng.blocks.EnhancedPacket(
            section=self.sbh,
            interface_id=self.interface_id,
            packet_data=packet_bytes,
            timestamp_high=(now_us & 0xFFFFFFFF00000000) >> 32,
            timestamp_low=(now_us & 0xFFFFFFFF),
            captured_len=len(packet_bytes),
            packet_len=len(packet_bytes),
            endianness="<",
            # options={
            #     'epb_flags': 0,
            #     'epb_tsresol': 6, # negative power of 10
            #     'epb_tsoffset': 0,
            #     'epb_len': len(packet_bytes),
            # },
        ))

    def stop(self, *args):
        logger.debug(f"graceful shutdown {args}")
        self.done = True



import serial
from collections.abc import Callable

from pymavlink import mavutil
from pymavlink.dialects.v20 import common as mavlink2


class MavSerial(mavutil.mavfile):
    '''Serial mavlink port'''
    def __init__(self, serial: serial.Serial, source_system=255, source_component=0, use_native=False, mavlink_version=2):
        self.baudrate = serial.baudrate
        self.name = serial.name
        self.autoreconnect = False
        self.force_connected = False
        self.port = serial
        super().__init__(None, serial.name, source_system=source_system, source_component=source_component, use_native=use_native)
        self.rtscts = False
        # correct mavlink serializer (self.mav) not to depend on environmental variables
        if mavlink_version == 1:
            from pymavlink.dialects.v10 import common as mavlink1
            self.mav = mavlink1.MAVLink(self, srcSystem=source_system, srcComponent=source_component)
            self.ParseError = mavlink1.MAVError
        else:
            from pymavlink.dialects.v20 import common as mavlink2
            self.mav = mavlink2.MAVLink(self, srcSystem=source_system, srcComponent=source_component)
            self.ParseError = mavlink2.MAVError
        # if you reach buffer size (4096) when using `loop://` then the program will hang unless you have the write timeout set
        if self.port.name is not None and "loop" in self.port.name and not self.port.write_timeout:
            self.port.write_timeout = 0.1

    def on_message_received(self, fn: Callable[[mavutil.mavfile, mavlink2.MAVLink_message], None]):
        self.message_hooks.append(fn)

    def set_baudrate(self, baudrate):
        '''set baudrate'''
        try:
            self.port.setBaudrate(baudrate)
        except Exception:
            # for pySerial 3.0, which doesn't have setBaudrate()
            self.port.baudrate = baudrate

    def close(self):
        self.port.close()

    def recv(self,n=None):
        if n is None:
            n = self.mav.bytes_needed()
        if self.fd is None:
            waiting = self.port.inWaiting()
            if waiting < n:
                n = waiting
        ret = self.port.read(n)
        return ret
    
    def __iter__(self):
        return self

    def __next__(self):
        '''handle mavlink packets'''
        while True:
            try:
                msg = self.recv_msg()
                if msg is None:
                    continue
                if msg.get_type() == "BAD_DATA":
                    continue
                return msg
            except serial.SerialException as e:
                raise StopIteration()

    def write(self, buf: bytes):
        try:
            return self.port.write(bytes(buf))
        except Exception:
            return -1

    def reset(self):
        raise RuntimeError("cannot reset serial port in MavSerial wrapper")

    def post_message(self, msg):
        # do not save messages into a in-memory queue
        pass